from mimetypes import guess_type
import random
import socket
from atavism.http11.content import FileContent
from atavism.http11.objects import HttpResponse
from atavism.http11.server import HttpServer

from atavism import __version__


class HLSServerError(Exception):
    pass


class HLSServer(HttpServer):
    """ Class that serves an HLSVideo via HTTP.
    """

    def __init__(self, video=None):
        HttpServer.__init__(self)
        self.video = video
        self.find_interface()

    def make_socket(self):
        failed = []
        while len(failed) < 5:
            port = random.randint(8100, 20000)
            try:
                sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                sock.setblocking(0)
                sock.settimeout(5)
                sock.bind((self.host, port))
                sock.listen(self.backlog)
                self.socket = sock
                self.port = port
                return
            except socket.error:
                pass
        raise HLSServerError("Unable to fid a suitable open port. Tried {}".format(",".join(failed)))

    @property
    def url(self):
        return "http://{}:{}{}".format(self.host, self.port, self.video.url)

    @property
    def content_type(self):
        return guess_type(self.video.url)[0]

    def find_interface(self):
        """ Find the local interface(s) that we will send via.
        :return: The interface to be used.
        """
        try:
            x = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            x.connect(('1.2.3.4', 56))
            interface = x.getsockname()[0]
            x.close()
            self.host = interface
            return True
        except socket.error as e:
            raise HLSServerError("Unable to find local interface.")

    def join(self):
        self.accept_thread.join(1.0)

    def handler(self, request):
        """ Create an HttpResponseObject from the request received. The server will call complete().
        :param request: The HttpRequest object to process.
        :return: The HttpResponse object.
        """
        self.logger.debug("Processing request for '%s'", request.path)
        resp = request.make_response()
        resp.add_headers({'Accept-Ranges': 'bytes',
                          'Server': 'atavism/{}'.format(__version__)})

        if request.method not in ('GET', 'HEAD'):
            resp.set_code(405)
            return resp

        rfn = self.video.find_file(request.path)
        if rfn is None:
            self.logger.info("Failed to find '%s'", request.path)
            resp.set_code(404)
            resp.set_content_type('text/html')
            resp.add_content("{} does not exist on this server.".format(request.path))
            return resp

        resp.set_content(FileContent(rfn))
        return resp



import argparse
from ipaddress import ip_address
import logging
import os
import sys
from atavism.devices import AirplayDevice, Chromecast, DeviceError
from atavism.dnssd import MDNSServiceDiscovery
from atavism.http import HLSServer
from atavism.video import find_ffmpeg, HLSVideo, SimpleVideo
from atavism import __version__

def main():
    parser = argparse.ArgumentParser(description='AppleTV Video Player')
    parser.add_argument('--find-devices', action='store_true',
                        help='Scan and report for AppleTV devices')
    parser.add_argument('--ip', help='IPv4 address of AppleTV to use')
    parser.add_argument('--send-direct', action='store_true',
                        help="Don't create an HLS stream, just send the file")
    parser.add_argument('--ffmpeg-binary-name', default='ffmpeg',
                        help='Name of ffmpeg binary to use')
    parser.add_argument('--ffmpeg-search-paths', help='Path(s) to search for ffmpeg binary')
    parser.add_argument('--hls-only', action='store_true', help='Just create an HLS stream')
    parser.add_argument('--chromecast', action='store_true', help='If an IP is supplied, is it for a Chromecast?')
    parser.add_argument('-v', nargs='*', help='Additional debug information')
    parser.add_argument('--version', action='store_true', help='Show version and exit')
    parser.add_argument('--log', help='Logfile to save output into')
    parser.add_argument('video', nargs='?', help="Video to stream")

    args = parser.parse_args()

    if args.version:
        print("atavism,  version {}".format(__version__))
        print("https://github.com/zathras777/atavism")
        sys.exit(0)

    if args.video is None and args.find_devices is False:
        print("You must supply a video filename unless --find-devices is used.")
        sys.exit(0)

    if args.video is not None and not os.path.exists(args.video):
        print("The file '{}' does not exist!".format(args.video))
        sys.exit(0)

    if args.hls_only and args.send_direct:
        print("You have used --send-direct and --hls-only. Sadly this app can't do both!")
        sys.exit(0)

    # Setup some simple logging
    logger = logging.getLogger()
    if args.v is not None:
        logger.setLevel(min(10, 30 - len(args.v) * 10))
    if args.log is not None:
        fh = logging.FileHandler(args.log)
        fh.setLevel(logger.level)
        logger.addHandler(fh)
    else:
        logger.addHandler(logging.StreamHandler())

    if not args.send_direct and args.find_devices is False:
        ffmpeg = find_ffmpeg(binary_name=args.ffmpeg_binary_name, paths=args.ffmpeg_search_paths)

    active_device = None
    devices = []

    if not args.hls_only:
        if args.find_devices is True or args.ip is None:
            print("Looking for devices...\n")
            dev_src = MDNSServiceDiscovery('_airplay._tcp.local', '_googlecast._tcp.local')
            if not dev_src.find_devices():
                print("Unable to locate any airplay devices. Exiting...")
                sys.exit(0)

            for ptr in dev_src.devices:

                if 'airplay' in ptr:
                    devices.append(AirplayDevice(dev_src.devices[ptr]))
                elif 'googlecast' in ptr:
                    devices.append(Chromecast(dev_src.devices[ptr]))

            for d in devices:
                print("    Found {}".format(d))
            print("\n Search complete.\n")

            if args.find_devices is True:
                for d in devices:
                    d.stop()
                sys.exit(0)

        elif args.ip is not None:
            if args.chromecast:
                devices = [Chromecast({'A': ip_address(args.ip.decode())})]
            else:
                devices = [AirplayDevice({'A': ip_address(args.ip.decode())})]

        if len(devices) > 1:
            for n in range(len(devices)):
                print("    {}. {}".format(n + 1, devices[n]))

            print("\nAs more than one device was found, please enter the number of the device to use:")
            while True:
                try:
                    input = raw_input
                except NameError:
                    pass
                num_str = input("Device to use (c to cancel): ")
                if num_str.lower() == 'c':
                    sys.exit(0)
                try:
                    num = int(num_str)
                except ValueError:
                    print("You need to enter a number!")
                    continue
                if 1 <= num <= len(devices):
                    active_device = devices[num - 1]
                    break
                print("Number must be between 1 and {}".format(len(devices)))
        else:
            active_device = devices[0]
    else:
        active_device = AirplayDevice()

    if not args.send_direct:
        video = HLSVideo(args.video)
        print("Creating the HLS stream... ")
        if not video.create_hls(active_device.width, active_device.height):
            print("Unable to create an HLS stream from '{}'".format(args.video))
            sys.exit(0)
        print("    done")

        if args.hls_only:
            video.cleanup = False
            print("\nHLS stream created in {}".format(video.directory))
            sys.exit(0)

        print("HLS stream created: {} segments...".format(video.segments))
    else:
        print("Getting video information...")
        video = SimpleVideo(args.video)
        print("    done")

    print("Duration: {} seconds\n".format(video.info.get('duration')))
    srv = HLSServer(video=video)
    srv.start()
    try:
        active_device.play_video(srv)
        while srv.running:
            try:
                srv.join()
            except KeyboardInterrupt:
                print("command_line: keyboard interrupt")
                break
    except DeviceError as e:
        print(e)
    srv.stop()



from datetime import datetime


def stripped_split(ss, c, n=-1):
    return [p.strip() for p in ss.split(c, n)]


class Cookie(object):
    def __init__(self, path=None, key=None, value=None, domain=None, expires=None, max_age=None, secure=False):
        self.path = path
        self.key = key
        self.value = value
        self.expires = expires
        self.domain = domain
        self.max_age = max_age
        self.secure = secure
        self.http_only = False

    def __eq__(self, other):
        if other.path != self.path or other.key != self.key or other.domain != self.domain:
            return False
        return True

    def __str__(self):
        base = ['{}={}'.format(self.key or '', self.value or '')]
        if self.path is not None:
            base.append("Path={}".format(self.path))
        if self.domain is not None:
            base.append("Domain={}".format(self.domain))
        if self.http_only:
            base.append('HttpOnly')

        return "; ".join(base)

    def set_expires(self, dtstr):
        self.expires = datetime.strptime(dtstr, "%a, %d-%b-%Y %H:%M:%S %Z")

    def as_header(self):
        return "{}={}".format(self.key, self.value)

    def is_relevant(self, _path=None):
        if self.expires is not None:
            if self.expires < datetime.utcnow():
                return False

        if _path is None:
            return False
        if self.path is None or _path == '/':
            return True

        if _path[:len(self.path)].lower() == self.path.lower():
            return True

        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other):
        if len(self.path) == len(other.path):
            return self.key < other.key
        return len(self.path) < len(other.path)

    def __gt__(self, other):
        return len(self.path) > len(other.path)


class CookieJar(object):
    def __init__(self):
        self.cookies = []

    def __len__(self):
        return len(self.cookies)

    def add_cookie(self, _cookie):
        for cc in self.cookies:
            if cc == _cookie:
                cc.value = _cookie.value
                return
        self.cookies.append(_cookie)

    def __getitem__(self, item):
        for c in self.cookies:
            if c.key == item:
                return c.value
        return None

    def get_cookie(self, item):
        for c in self.cookies:
            if c.key == item:
                return c
        return None

    def parse_set_cookie(self, hdr_string):
        if '=' not in hdr_string:
            return
        parts = stripped_split(hdr_string, ';')
        c = Cookie()
        c.key, c.value = stripped_split(parts[0], '=', 1)
        for p in parts[1:]:
            if p == 'HttpOnly':
                c.http_only = True
                continue

            k, v = stripped_split(p, '=', 1)
            if k.lower() == 'expires':
                c.set_expires(v)
            else:
                setattr(c, k.lower(), v)
        self.add_cookie(c)

    def check_cookies(self, http_obj):
        cookies = http_obj.get('set-cookie')
        if cookies is None:
            return
        for c_str in cookies:
            self.parse_set_cookie(c_str)

    def get_cookies(self, _path):
        matched = []
        for c in self.cookies:
            if c.is_relevant(_path):
                matched.append(c)
        if len(matched) == 0:
            return None
        return '; '.join([c.as_header() for c in sorted(matched)])



import json
import logging
from random import randint
from select import select
from struct import pack, unpack, unpack_from
import socket
import ssl
from threading import Thread, Event

try:
    from queue import Queue, Empty
except ImportError:
    from Queue import Queue, Empty


CONNECTION_NS = "urn:x-cast:com.google.cast.tp.connection"
HEARTBEAT_NS = 'urn:x-cast:com.google.cast.tp.heartbeat'
RECEIVER_NS = 'urn:x-cast:com.google.cast.receiver'
AUTH_NS = 'urn:x-cast:com.google.cast.tp.deviceauth'
MEDIA_NS = 'urn:x-cast:com.google.cast.media'

PLATFORM_DEST = 'receiver-0'


class ProtoBuff(object):
    PROTOCOL_VERSION = 0
    TYPE_ENUM = 0
    TYPE_STRING = 2
    TYPE_BYTES = TYPE_STRING
    TYPE_NAMES = { 0: 'String', 1: 'Binary' }

    def __init__(self, **kwargs):
        self.protocol = kwargs.get('protocol', self.PROTOCOL_VERSION)
        self.source_id = kwargs.get('source_id', 'source-0')
        self.destination_id = kwargs.get('destination_id', PLATFORM_DEST)
        self.namespace = kwargs.get('namespace', CONNECTION_NS)
        self.type = kwargs.get('type', 0)
        self.data = kwargs.get('data', '')
        if 'json' in kwargs:
            self.from_json(kwargs['json'])
        if 'msg' in kwargs:
            self.from_string(kwargs['msg'], kwargs.get('msg_len', len(kwargs['msg'])))

    @staticmethod
    def _pack_type(field_id, t):
        return (field_id << 3) | t

    @staticmethod
    def _unpack_type(val):
        return val >> 3, val & 0x7

    @staticmethod
    def _data_length(s):
        x = b""
        l = len(s)
        while (l > 0x7F):
            x += pack("B", l & 0x7F | 0x80)
            l >>= 7
        x += pack("B", l & 0x7F)
        return x

    @staticmethod
    def _unpack_varint(bytes):
        """ Convert a varint to an integer.
        :param bytes: Bytes containing the varint.
        :return: integer
        """
        value = 0
        base = 1
        rd = 0
        for raw_byte in bytes:
            val_byte = ord(raw_byte) if type(raw_byte) != int else raw_byte
            rd += 1
            value += (val_byte & 0x7f) * base
            if (val_byte & 0x80):
                # The MSB was set; increase the base and iterate again, continuing
                # to calculate the value.
                base *= 128
            else:
                break
        return value, rd

    def _pack_string(self, n, the_str):
        sl = len(the_str)
        return pack(">BB%ds" % sl, self._pack_type(n, self.TYPE_STRING), sl, the_str.encode())

    @staticmethod
    def _unpack_string(buff):
        slen = unpack_from(">B", buff, 0)[0]
        ss = unpack(">%ds" % slen, buff[1:1+slen])[0]
        return ss, slen + 1

    def as_string(self):
        _msg = pack(">BB", self._pack_type(1, self.TYPE_ENUM), self.PROTOCOL_VERSION)
        _msg += self._pack_string(2, self.source_id)
        _msg += self._pack_string(3, self.destination_id)
        _msg += self._pack_string(4, self.namespace)
        _msg += pack(">BB", self._pack_type(5, self.TYPE_ENUM), self.type)
        _msg += pack(">B", self._pack_type(6, self.TYPE_BYTES))
        _msg_len = self._data_length(self.data)
        _msg += pack(">%ds" % len(_msg_len), _msg_len)
        _msg += pack(">%ds" % len(self.data), self.data.encode())
        return pack(">I%ds" % (len(_msg)), len(_msg), _msg)

    def from_string(self, bytes, blen):
        pos = 0
        while pos < blen:
            a = unpack_from(">B", bytes, pos)[0]
            n, ct = self._unpack_type(a)
            pos += 1
            if n == 1:
                self.protocol = unpack_from(">B", bytes, pos)[0]
                pos += 1
            elif n == 2:
                self.source_id, rd = self._unpack_string(bytes[pos:])
                pos += rd
            elif n == 3:
                self.destination_id, rd = self._unpack_string(bytes[pos:])
                pos += rd
            elif n == 4:
                self.namespace, rd = self._unpack_string(bytes[pos:])
                pos += rd
            elif n == 5:
                self.type = unpack_from(">B", bytes, pos)[0]
                pos += 1
            elif n == 6: # utf-8 payload
                slen, rd = self._unpack_varint(bytes[pos:])
                pos += rd
                self.data = unpack(">%ds" % slen, bytes[pos:pos + slen])[0]
                pos += slen
            elif n == 7: # binary payload
                slen, rd = self._unpack_varint(bytes[pos:])
                pos += rd
                self.data = unpack_from(">%dB" % slen, bytes, pos)[0]
                pos += slen
            else:
                print("n = {}".format(n))
                break

    def as_json(self):
        if self.type == 0:
            return json.loads(self.data.decode())

    def from_json(self, json_data):
        self.data = json.dumps(json_data)

    def dump(self):
        """ Dump info to stdout... """
        print("  Protocol Version:   {}".format(self.protocol))
        print("  Source ID:          {}".format(self.source_id))
        print("  Desination ID:      {}".format(self.destination_id))
        print("  Namespace:          {}".format(self.namespace))
        print("  Type:               {} [{}]".format(self.TYPE_NAMES[self.type], self.type))
        print("  Data:               {}".format(self.data))


class ChromecastClient(object):
    """ Class to communicate with a Chromecast device.

        In order to communicate a virtual connection is required, estblished by using the connect() function.

        Once established the Chromecast will send a PING request, which must be answered with a PONG reply
        or the connection will be closed.


        Useful links:

         - https://developers.google.com/cast/docs/reference/messages

    """
    DEFAULT_MEDIA_APP = 'CC1AD845'

    def __init__(self, host, port=8009):
        self.logger = logging.getLogger()
        self.req_id = None
        self.volume = 0
        self.muted = False
        self.stadby = True
        self.active = False
        self.host = host
        self.port = port

        self.sessions = {}
        self.sessions_updated = Event()
        self.available_apps = set()
        self.apps_available = Event()

        self.connected = False
        self.socket = None

        self.input = Queue()
        self.output = Queue()

        self.running = False

        self.comm_thread = None
        self.switch_thread = None

        self.heartbeat = HeartbeatReceiver(self)

        self.responses = {}
        self.events = {}

    def start(self):
        if self.running is True:
            return

        self.socket = ssl.wrap_socket(socket.socket())
        self.socket.settimeout(10)
        try:
            self.socket.connect((str(self.host), self.port))
        except socket.gaierror as e:
            self.logger.warning("Unable to connect to device. %s", e)
            return

        self.running = True
        self.switch_thread = Thread(target=self.switchboard)
        self.switch_thread.daemon = True
        self.switch_thread.start()

        self.comm_thread = Thread(target=self.communicator)
        self.comm_thread.daemon = True
        self.comm_thread.start()

        self.connect()
        self.get_app_availability(self.DEFAULT_MEDIA_APP)

    def stop(self):
        self.running = False

    def register_event(self, request_id):
        ev = Event()
        self.events[request_id] = ev
        return ev

    def unregister_event(self, request_id):
        if request_id not in self.events:
            return
        del self.events[request_id]

    def communicator(self):
        _buffer = b''
        while self.running:
            w = [self.socket] if self.output.not_empty else []
            _r, _w, _e = select([self.socket], w, [self.socket], 10)

            if len(_e):
                break

            if len(_r) > 0:
                _buffer += self.socket.recv(2048)

            if len(_w) > 0:
                try:
                    pb = self.output.get_nowait()
#                   if pb.namespace != HEARTBEAT_NS:
#                       print(">>>>>>>>>>>>>>")
#                       pb.dump()
                    wl = self.socket.write(pb.as_string())
                    self.output.task_done()
                except Empty:
                    pass

            if len(_buffer) >= 4:
                plen = unpack(">I", _buffer[:4])[0]
                if len(_buffer) >= plen + 4:
                    self.accept_message(_buffer[4:4 + plen], plen)
                    _buffer = _buffer[4 + plen:]

    def switchboard(self):
        while self.running:
            pb = self.input.get()
            self.input.task_done()

            if pb.namespace == HEARTBEAT_NS:
                self.heartbeat.process_message(pb)
                continue

#            print("<<<<<<<<<<<<<<")
#            pb.dump()

            msg = pb.as_json()
            if pb.namespace == CONNECTION_NS:
                if msg.get('type') == 'CLOSE':
                    if pb.source_id in self.sessions:
                        self.sessions[pb.source_id].connected = False
                    continue
            if pb.namespace == MEDIA_NS and pb.destination_id == '*':
                sess = self.sessions.get(pb.source_id)
                if sess is not None:
                    sess.update_media_status(pb.as_json())

            if msg.get('requestId') in self.events:
                self.responses[msg['requestId']] = pb
                self.events[msg['requestId']].set()

    def put_and_wait(self, pb, payload, timeout=10):
        req_id = self.request_id
        payload['requestId'] = req_id
        ev = self.register_event(req_id)
        pb.from_json(payload)
        self.output.put(pb)
        rv = ev.wait(timeout)
        self.unregister_event(req_id)
        if rv and req_id in self.responses:
            rv = self.responses[req_id]
            del self.responses[req_id]
        return rv

    def accept_message(self, bytes, blen):
        pb = ProtoBuff(msg=bytes, msg_len=blen)
        self.input.put(pb)

    @property
    def request_id(self):
        if self.req_id is None:
            self.req_id = randint(1000000, 80000000)
        self.req_id += 1
        return self.req_id

    def stop_apps(self):
        for sess in self.sessions.values():
            self.put_and_wait(ProtoBuff(namespace=RECEIVER_NS), {'type': 'STOP', 'sessionId': sess.session_id})
            sess.disconnect()
            del self.sessions[sess.session_id]

    def connect(self):
        self.output.put(ProtoBuff(data="{\"type\":\"CONNECT\"}"))

    def get_status(self):
        self.put_and_wait(ProtoBuff(namespace=RECEIVER_NS), {'type': 'GET_STATUS'})

    def get_app_availability(self, *apps):
        """ Enquire whether one or more apps are available on the Chromecast.
            We use the cached list wherever possible and block until we have an
            answer.
        :param apps: One or more app names or ids to search for.
        :return: Dict with each app requested as a True/False.
        """
        if len(apps) == 0:
            return
        toask = [a for a in apps if a not in self.available_apps]
        if len(toask) > 0:
            payload = {'type': 'GET_APP_AVAILABILITY',
                       'appId': toask}
            resp = self.put_and_wait(ProtoBuff(namespace=RECEIVER_NS), payload)
            if resp is not False:
                data = resp.as_json()
                for app in data.get('availability', []):
                    if data['availability'][app] == 'APP_AVAILABLE':
                        self.available_apps.add(app)
        return {a: a in self.available_apps for a in apps}

    def launch_app(self, app_id=None, block=True):
        payload = {'type': 'LAUNCH',
                   'appId': app_id or self.DEFAULT_MEDIA_APP}
        resp = self.put_and_wait(ProtoBuff(namespace=RECEIVER_NS), payload)
        if resp is False:
            return None
        app_data = resp.as_json().get('status').get('applications')[0]
        if 'transportId' not in app_data:
            print("The app has been loaded but does not use the cast API.")
            return None
        sess = ChromecastSession(self, app_data)
        self.sessions[app_data['transportId']] = sess
        return sess

    def update_status(self, pb):
        msg = pb.as_json()
        self.volume = msg['status']['volume']['level']
        self.muted = msg['status']['volume']['muted']
        self.standby = msg['status']['isStandBy']
        self.active = msg['status'].get('isActiveInput', False)

        if 'applications' in msg['status']:
            for app in msg['status']['applications']:
                if app['sessionId'] not in self.sessions:
                    self.sessions[app['sessionId']] = ChromecastSession(self, app)
                else:
                    self.sessions[app['sessionId']].update(app)
                self.sessions_updated.set()


class ChromecastSession(object):
    """ Class to contain details of a session on a Chromecast.
    """
    def __init__(self, client, data):
        self.client = client
        self.app_id = data['appId']
        self.display_name = data['displayName']
        self.namespaces = [ns['name'] for ns in data.get('namespaces', [])]
        self.session_id = data['sessionId']
        self.status = data['statusText']
        self.transport_id = data.get('transportId', None)
        self.connected = False

        self.media_loaded = False
        self.media_position = 0
        self.media_status = ''
        self.media_session_id = 0
        self.media_finished = False

    @property
    def uses_cast_api(self):
        return self.transport_id is not None

    def connect(self):
        if self.transport_id is None:
            return
        self.client.output.put(ProtoBuff(namespace=CONNECTION_NS, destination_id=self.transport_id, json={'type': 'CONNECT'}))
        ck = self.client.put_and_wait(ProtoBuff(namespace=MEDIA_NS, destination_id=self.transport_id),
                                                            {'type': 'GET_STATUS'})
        self.connected = ck is not False

    def disconnect(self):
        if not self.connected:
            return
        self.client.output.put(ProtoBuff(namespace=CONNECTION_NS, destination_id=self.transport_id, json={'type': 'CLOSE'}))

    def get_media_status(self):
        if not self.connected or not self.media_loaded:
            return
        update = self.client.put_and_wait(ProtoBuff(namespace=MEDIA_NS, destination_id=self.transport_id), {'type': 'GET_STATUS'})
        if update is not False:
            self.update_media_status(update.as_json())

    def update_media_status(self, data):
        """ This can be called from the parent (if the update was broadcast) or from get_media_status().
        :param data: The update data.
        """
        if data.get('type') != 'MEDIA_STATUS':
            return
        status = data.get('status', {})
        if isinstance(status, list):
            if len(status) == 0:
                return
            status = status[0]
        self.media_session_id = status.get('mediaSessionId')
        self.media_status = status.get('playerState')
        self.media_position = status.get('currentTime')
        if 'idleReason' in status:
            self.media_finished = True

    def get_status(self):
        update = self.client.put_and_wait(ProtoBuff(namespace=self.namespaces[0], destination_id=elf.transport_id),
                                 {'type': 'GET_STATUS', 'mediaSessionId': self.session_id})
        print(update.as_json())

    def load_movie(self, url, ct, duration=None):
        if not self.connected:
            return
        payload = {'type': 'LOAD',
                   'media': {'contentId': url, 'contentType': ct, 'streamType': 'BUFFERING'},
                   'autoplay': False
                   }
        if duration is not None:
            payload['media']['duration'] = duration
        resp = self.client.put_and_wait(ProtoBuff(namespace=self.namespaces[0],
                                                   destination_id=self.transport_id), payload, 15)
        if resp is False:
            print("Chromecast is unable to load media?")
            self.media_loaded = False
            return False
        self.media_loaded = True
        self.get_media_status()
        return True

    def play_media(self):
        if not self.media_loaded:
            return
        payload = {'type': 'PLAY', 'mediaSessionId': self.media_session_id}
        resp = self.client.put_and_wait(ProtoBuff(namespace=self.namespaces[0],
                                                  destination_id=self.transport_id), payload)
        return resp is not False


class ChromecastReceiver(object):
    def __init__(self, client, **kwargs):
        self.client = client
        self.namespace = kwargs.get('namespace')
        self.source = kwargs.get('source_id')
        self.dest = kwargs.get('destination_id')
        self.namespace = kwargs.get('namespace')

    def update_from_message(self, pb):
        if self.source is None:
            self.source = pb.source_id
        if self.dest is None:
            self.dest = pb.destination_id
        if self.namespace is None:
            self.namespace = pb.namespace

    def process_message(self, pb):
        self.update_from_message(pb)
        msg = pb.as_json()


class HeartbeatReceiver(ChromecastReceiver):
    def __init__(self, client, **kwargs):
        ChromecastReceiver.__init__(self, client, **kwargs)

    def process_message(self, pb):
        self.update_from_message(pb)

        msg = pb.as_json()
        if 'type' in msg and msg['type'] == 'PING':
            self.client.output.put(ProtoBuff(source_id=self.source,
                                             destination_id=self.dest,
                                             namespace=self.namespace,
                                             json={'type': 'PONG'}))



from errno import EPERM, EACCES
from os import path
import os
import sys
import subprocess
from tempfile import mkdtemp
import re
import math


def find_ffmpeg(binary_name='ffmpeg', skip_list=None, paths=None, silent=False):
    """ Function to find and return the full path to a suitable ffmpeg binary.
        If no suitable ffmpeg binary is found, a RuntimeError exception is raised.
    :param binary_name: The name of the binary to search for. Default is ffmpeg
    :param skip_list: List of binary files to skip.
    :param paths: Additional system paths to include in search.
    :return: Path to ffmpeg binary.
    """
    found = False
    if skip_list is None:
        skip_list = []
    os_paths = ['/usr/bin', '/usr/local/bin']
    if paths is not None:
        if isinstance(paths, (list, tuple)):
            for p in reversed(paths):
                os_paths.insert(0, p)
        else:
            os_paths.insert(0, paths)
    for p in os_paths:
        poss = os.path.join(p, binary_name)
        if os.path.exists(poss) and os.path.isfile(poss) and poss not in skip_list:
            found = True
            break

    if not found:
        raise RuntimeError("\nNo binary '{}' has been found. Paths tried: {}".
                           format(binary_name, ','.join(os_paths)))

    # Check for hls support
    try:
        p = subprocess.Popen([poss, '-protocols'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        out, err = p.communicate()
    except (IOError, OSError) as e:
        if e.errno in (EPERM, EACCES):
            if not silent:
                print("\n'{}' is not an executable file.".format(poss))
            found = False

    if not found:
        skip_list.append(poss)
        return find_ffmpeg(binary_name=binary_name, skip_list=skip_list)

    if b'hls' not in out:
        if not silent:
            print("HLS was not found in the protocols list for '{}'.\n"\
                  "Without HLS support we can't use this {}.".format(poss, binary_name))
        skip_list.append(poss)
        return find_ffmpeg(binary_name=binary_name, skip_list=skip_list)

    # Check for libx264
    p = subprocess.Popen([poss, '-encoders'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    out, err = p.communicate()
    if b'libx264' not in out:
        if not silent:
            print("libx264 was not found in the encoders list for '{}'.\n"\
                  "As we require libx264, we can't use this {}.".format(poss, binary_name))
        skip_list.append(poss)
        return find_ffmpeg(binary_name=binary_name, skip_list=skip_list)

    return poss


class BaseVideo(object):
    def __init__(self, source, ffmpeg=None):
        self.source = source
        self.directory = path.dirname(source)
        if self.directory.startswith('~'):
            self.directory = path.expanduser(self.directory)
        self.directory = path.abspath(self.directory)
        self.ffmpeg = ffmpeg or find_ffmpeg()
        self.info = {}
        self.streams = []
        self.video_stream = None
        self.get_video_information()

    @property
    def url(self):
        return "/{}".format(os.path.basename(self.source))

    def find_file(self, url):
        poss_fn = path.join(self.directory, url[1:] if url.startswith('/') else url)
        if not os.path.exists(poss_fn):
            return None
        return poss_fn

    def _execute_ffmpeg(self, *args):
        cmd_args = [self.ffmpeg, '-i', self.source]
        cmd_args.extend(*args)
        p = subprocess.Popen(cmd_args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        return p.communicate()

    def get_video_information(self):
        ignored, data = self._execute_ffmpeg([])
        for input in data.split(b'\nInput')[1:]:
            for l in [ln.strip() for ln in input.split(b'\n')]:
                if l.startswith('Duration'):
                    dur, ignored = l[10:].split(',', 1)
                    parts = [float(p) for p in dur.split(':')]
                    self.info['duration'] = parts[0] * 3600 + parts[1] * 60 + parts[2]

                if not l.startswith(b'Stream'):
                    continue
                info = re.match(b'^Stream #([0-9]\:[0-9])\(?([A-Za-z]{2,})?\)?: ([A-Za-z]+):', l)
                if info is None:
                    continue
                sinfo = {'type': info.group(3), 'n': info.group(1), 'lang': info.group(2)}
                parts = b'.'.join(l.split(b': ')[2:]).split(b', ')

                if info.group(3) == b'Video':
                    if self.video_stream is None:
                        self.video_stream = sinfo
                    for p in parts:
                        if b'fps' in p:
                            sinfo['fps'] = float(p.replace(b'fps', b'').strip())
                        elif re.search(b'[0-9]{2,}x[0-9]{2,}', p):
                            sz = re.search(b'([0-9]{2,})x([0-9]{2,})', p)
                            sinfo['width'] = int(sz.group(1))
                            sinfo['height'] = int(sz.group(2))
                self.streams.append(sinfo)

    def video_width(self):
        if self.video_stream is None:
            return -1.0
        return float(self.video_stream.get('width', -1))

    def video_height(self):
        if self.video_stream is None:
            return -1.0
        return float(self.video_stream.get('height', -1))


class SimpleVideo(BaseVideo):
    def __init__(self, source):
        BaseVideo.__init__(self, source)


class HLSVideo(BaseVideo):
    """ We will attempt to create a temporary directory to contain the segments of an HLS
        stream. The files are removed when the instance that created them is deleted.
    """
    def __init__(self, source, tmp_base=None, ffmpeg=None):
        BaseVideo.__init__(self, source, ffmpeg)
        self.cleanup = True
        self.fn = os.path.splitext(os.path.basename(source))[0] + '.m3u8'
        self.directory = mkdtemp(dir=tmp_base or '/tmp')
        self.hls_filename = os.path.join(self.directory, self.fn)
        self.meta = {}
        self.duration_data = {}
        self.hls_time = 10
        self.segments = 0

    @property
    def url(self):
        return "/{}".format(self.fn)

    def __del__(self):
        if self.cleanup:
#            print("Removing directory {}".format(self.directory))
            for f in os.listdir(self.directory):
                os.unlink(os.path.join(self.directory, f))
            os.rmdir(self.directory)

    def create_hls(self, max_width=-1, max_height=-1):
        opts = []
        if self.needs_resize(max_width, max_height):
            opts.extend(['-vf', 'scale={}:{}'.format(*self._resized(max_width, max_height))])

        output, err = self._hls_command(opts)
        self.segments = len(os.listdir(self.directory)) - 1
        if self.segments > 0:
            return True
        print(output)
        print(err)

        return False

    def has_audio(self):
        for s in self.streams:
            if s.get('type') == b'Audio':
                return True
        return False

    def audio_streams(self):
        n = 0
        for s in self.streams:
            if s.get('type') == b'Audio':
                n += 1
        return n

    def needs_resize(self, max_width=-1, max_height=-1):
        if max_width == -1 and max_height == -1:
            return False
        if max_width != -1 and self.video_width() > max_width:
            return True
        if max_height != -1 and self.video_height() > max_height:
            return True
        return False

    def _resized(self, max_width, max_height):
        rw = max_width / self.video_width() if max_width != -1 else 1.0
        rh = max_height / self.video_height() if max_height != -1 else 1.0
        ratio = min(rw, rh)
        w = int(math.floor(self.video_width() * ratio))
        h = int(math.floor(self.video_height() * ratio))
        if w % 2 != 0:
            w -= 1
        if h % 2 != 0:
            h -= 1
        return w, h

    def _hls_command(self, *opts):
        args = ['-hls_time', str(self.hls_time), '-hls_list_size', '0', '-f', 'hls']
        args.extend(*opts)
        args += [self.hls_filename]
        return self._execute_ffmpeg(args)



import logging
import socket
import ssl
from threading import Thread, Event
from time import sleep, time
import re
import math
from struct import pack
from atavism.chromecast import ChromecastClient
from atavism.http11.client import HttpClient


class DeviceError(Exception):
    pass


def show_progress(current, duration):
    interval = 2
    if current > 0:
        n = (current / duration) * 100
        bar = "#" * int(math.floor(n / 2.0))
    elif current == duration:
        bar = ' Completed'
        n = 100.0
        interval = 0
    else:
        bar = ' Buffering...'
        n = 0.0
    print('\r   Playback: [%-50s] %7.03f%%' % (bar, n))
    return interval


class AirplayDevice:
    SCRUB_RE = re.compile(b"([a-z]+)\: ([0-9\.]+)")
    UPDATE_INTERVAL = 5

    def __init__(self, device=None):
        self.name = None
        self.info = None
        self.output = (-1, -1)
        if device is None:
            return

        self.ptr = device.get('PTR')
        self.host = device.get('A')
        self.host6 = device.get('AAAA')
        self.port = device.get('port', 7000)

        self.http = HttpClient(self.host, self.port)

        self.logger = logging.getLogger(__name__)
        self.logger.setLevel(logging.getLogger().level)

        if self.ptr is not None:
            self.name = self.ptr.split('.')[0]
            self.get_info()

    @property
    def width(self):
        return self.output[0]

    @property
    def height(self):
        return self.output[1]

    def get_info(self):
        if self.host is None and self.host6 is None:
            raise DeviceError("Unable to get device information as no IPv4 or IPv6 address available?")

        self.info = self.http.simple_request('/server-info')
        if self.info is None:
            return
        self.output = (1920, 1080) if '3' in self.info['model'] else (1280, 720)

    def __str__(self):
        if self.info is None:
            return "Unknown device @ {}:{}".format(self.host, self.port)
        return "{} [{}]  {}x{} @ {}:{}".format(self.name or 'Unknown',
                                               self.info.get('model', 'Unknown'),
                                               self.output[0], self.output[1],
                                               self.host, self.port)

    def features(self):
        if self.info is None or 'features' not in self.info:
            return 'Unknown'
        supported = []
        f_list = ['Video', 'Photo', 'VideoFairPlay', 'VideoVolumeControl', 'VideoHTTPLiveStreams',
                  'Slideshow', '6', 'Screen', 'ScreenRotate', 'Audio', 'AudioRedundant',
                  'FPSAPv2pt5_AES_GCM', 'PhotoCaching', '12', '13', '14', '15', '16', '17',
                  '18', '19', '20', '21', '22', '23', '24', '25', '26', '27', '28', '29',
                  '30', '31']
        for i in range(len(f_list)):
            if self.info['features'] & 1 << i != 0:
                supported.append(f_list[i])
        return ",".join(supported)

    @property
    def supports_streaming(self):
        if self.info is None or 'features' not in self.info:
            return False
        return self.info.get('features', 0) & 1 << 4 != 0

    def stop_video(self):
        resp = self.http.post_data('/stop')
        return resp.code == 200

    def play_video(self, video_srv):
        self.stop_video()

        pdata = {'Content-Location': video_srv.url, 'Start-Position': 0}
        resp = self.http.post_data('/play', data=pdata, ct='text/parameters')
        if resp is None or resp.code != 200:
            self.logger.error("Unable to play the video. Code returned was %s", resp.code if resp is not None else None)
            raise DeviceError("Unable to play video")

        current, duration = -1, 0
        try:
            while duration <= duration:
                interval = show_progress(current, duration)
                if interval == 0:
                    break
                sleep(interval)
                current, duration = self.get_position()
        except KeyboardInterrupt:
            print("\r\nStopping...")

        print('\r\n')
        video_srv.stop()

    def get_position(self):
        data = self.http.simple_request('/scrub')
        if data is None:
            return 0.0, 0.0

        def getfloat(key):
            return float(data.get(key, '0.0'))

        return getfloat(b'position'), getfloat(b'duration')

    def stop(self):
        data = self.http.post_data('/stop')
        if data.code == 200:
            return True
        return False


class Chromecast(object):
    def __init__(self, device=None):
        if device is None:
            return

        self.width = 1920
        self.height = 1080

        self.info = {}
        self.txt = device.get('TXT')
        self.output = (-1, -1)
        self.ptr = device.get('PTR')
        self.host = device.get('A')
        self.host6 = device.get('AAAA')
        srv = device.get('SRV', {})
        self.port = srv.get('port', 8009)
        self.name = srv.get('name')

        self.http = HttpClient(self.host, self.port)
        self.dial = HttpClient(self.host, 8008)
        if self.ptr is not None:
            self.get_info()

        self.client = ChromecastClient(self.host, self.port)

    def __str__(self):
        return "Chromecast: {} @ {}".format(self.name, self.host)

    def get_info(self):
        UPNP_NS = "{urn:schemas-upnp-org:device-1-0}"
        xml = self.dial.simple_request('/ssdp/device-desc.xml')
        if xml is None:
            return

        def get_xpath(_tree, _node, raw=False):
            _xpath = './/' + '/'.join(['*[local-name()="{}"]'.format(p) for p in _node.split('/')])
            matches = _tree.xpath(_xpath)
            if raw:
                return matches
            if len(matches) > 0:
                return matches[0].text
            return ''

        info = {}
        info['URLBase'] = get_xpath(xml, 'URLBase')
        info['api_version'] = "{}.{}".format(get_xpath(xml, 'specVersion/major'),
                                             get_xpath(xml, 'specVersion/minor'))
        info['deviceType'] = get_xpath(xml, 'device/deviceType')
        info['friendlyName'] = get_xpath(xml, 'device/friendlyName')
        info['manufacturer'] = get_xpath(xml, 'device/manufacturer')
        info['modelName'] = get_xpath(xml, 'device/modelName')
        info['UDN'] = get_xpath(xml, 'device/UDN')
        self.info = info

        svcs = []
        for svc in get_xpath(xml, 'serviceList/service', True):
            vals = ('serviceType', 'serviceId', 'controlURL', 'eventSubURL', 'SCPDURL')
            svcs.append({k: get_xpath(svc, k) for k in vals})
        self.services = svcs

    def reboot(self):
        """ Ask the Chromecast to reboot.
        :return: None.
        """
        self.dial.post_data("/setup/reboot", data='{"params": "now"}', ct='application/json')

    def stop(self):
        if self.client.running is True:
            return
        self.client.stop()

    def stop_video(self):
        if self.client.running:
            self.client.stop_apps()
        return True

    def play_video(self, video_srv):
        self.client.start()
        if not self.client.running:
            return

        ck = self.client.get_app_availability(self.client.DEFAULT_MEDIA_APP)
        if not ck.get(self.client.DEFAULT_MEDIA_APP, False):
            return
        self.client.stop_apps()
        sess = self.client.launch_app()
        if sess is None:
            self.logger.warning("Unable to get a media player session.")
            video_srv.stop()
            return

        sess.connect()
        if not sess.connected:
            self.logger.warning("Unable to connect to the media player session.")
            video_srv.stop()
            return
        duration = video_srv.video.info.get('duration')
        ck = sess.load_movie(video_srv.url, video_srv.content_type, duration)
        if ck:
            sess.play_media()

        while not sess.media_finished:
            try:
                sess.get_media_status()
                sleep(0.25)
                if not sess.media_finished:
                    interval = show_progress(sess.media_position, duration)
                    sleep(interval)
            except KeyboardInterrupt:
                print("\nStopping....\n")
                break

        print("Finished.")
        video_srv.stop()



class Range(object):
    def __init__(self, tpl):
        self.start = int(tpl[0]) if tpl[0] is not None and tpl[0] != b'' else None
        self.end = int(tpl[1]) if tpl[1] is not None and tpl[1] != b'' else None
        if self.start is None and self.end is not None and self.end > 0:
            self.end *= -1

    def __str__(self):
        return "Byte Range: {} - {}".format(self.start, self.end)

    def __len__(self, cl=0):
        if self.start is None and self.end is not None:
            return self.end * -1
        elif self.end is None:
            return cl - self.start
        return self.end - self.start + 1

    def header(self):
        r = ''
        if self.start is not None:
            r += '{}-'.format(self.start)
        if self.end is not None:
            r += "{}".format(self.end)
        return r

    def from_content(self, content):
        """ Try and get the range from the supplied content. If it isn't possible,
            return None.
        :param content: The content stream to extract the range from.
        :return: The extracted content or None.
        """
        csz = len(content)
        if self.end is not None:
            if self.end < 0 and csz < self.end * -1:
                print("not big enough")
                return None
            if self.end > csz:
                print("end > content length")
                return None
        else:
            if self.start > csz:
                print("start > content length")
                return None
        if self.end is None:
            return content[self.start:]
        elif self.start is None:
            return content[self.end:]
        return content[self.start: self.end + 1]

    def absolutes(self, clen):
        start = self.start
        if self.start is None:
            if self.end < 0:
                return clen + self.end, clen - 1
            start = 0
        end = clen - 1
        if self.end is not None:
            if self.end < 0:
                end = clen + self.end - 1
            else:
                end = self.end - 1
        if end < start:
            end = start
        return start, end

    def absolute_range(self, clen):
        start, end = self.absolutes(clen)
        return "{}-{}/{}".format(start, end, clen)



import re
from atavism.http11.content import Content
from atavism.http11.headers import Headers
from atavism.http11.range import Range


class BaseHttp(object):
    """ Base class for other HTTP transactional classes. This class tries
        to provide the core functionality for various classes.
    """
    RANGE_re = re.compile(r"([0-9]+)?-([0-9]+)?,?")

    def __init__(self):
        self.path = ''
        self.http = 'HTTP/1.1'

        self.header = Headers()
        self._content = Content()

        self.headers_sent = False
        self.headers_only = False
        self.ranges = []

    def __len__(self):
        return len(self._content)

    def get(self, key, default=None):
        return self.header.get(key, default)

    def add_header(self, key, val):
        self.header.add_header(key, val)

    def add_headers(self, hdr_dict):
        self.header.add_headers(hdr_dict)

    ### Connection Information
    @property
    def is_keepalive(self):
        return False if self.header.get('connection', '').lower() == 'close' else True

    def send_complete(self):
        ct = self._content
        while ct._next is not None:
            ct = ct._next
        if self.headers_sent and ct.send_complete:
            return True
        return False

    def next_output(self):
        data = b''
        if not self.headers_sent:
            data += str(self.header).encode()
            self.headers_sent = True
        if self.headers_only:
            self._content.finished = True
            return data

        data += self._content.next(len(data))
        return data

    ### Ranges
    def parse_ranges(self, key):
        """ Parse a range request into individual byte ranges. """
        self.ranges = []
        poss = self.get(key)
        if poss is None or poss.lower() == 'none':
            return
        if not poss.startswith("bytes="):
            return
        matches = self.RANGE_re.findall(poss[6:])
        for m in matches:
            self.ranges.append(Range(m))

    def add_range(self, start=None, end=None):
        if start is None and end is None:
            return
        self.ranges.append(Range((start, end)))

    def has_ranges(self):
        return len(self.ranges) > 0

    def set_ranges(self, ranges):
        self.ranges = ranges

    ### Content

    @property
    def content(self):
        return self._content.content

    def decoded_content(self):
        return self._content.decoded_content()

    def set_content(self, cntnt_obj):
        self._content = cntnt_obj

    def set_content_type(self, ct):
        self._content.content_type = ct

    def read_content(self, cntnt):
        r = 0
        if not self.header.finished:
            r = self.header.read_content(cntnt)
            if self.header.finished:
                self._update_content()
        r += self._content.read_content(cntnt[r:])
        return r

    def _update_content(self):
        self._content.content_type = self.header.get('content-type')
        te = self.get('transfer-encoding')
        cl = self.get('content-length')
        # Set how we will detect the length of content...
        if te is not None and te.lower() == 'chunked':
            self._content.content_sz = 'chunked'
        elif cl is not None:
            self._content.content_sz = cl
        else:
            self._content.content_sz = None

        rngs = self.get('range')
        if rngs is not None:
            self.parse_ranges('range')

        ce = self.get('content-encoding')
        if ce is not None and ce.lower() != 'identity':
            self._content.set_compression(ce)

    def set_compression(self, method):
        self._content.set_compression(method)

    def _complete(self):
        """ Record that the creation of a response/request is complete. """
        self._content.finished = True
        self._content.compress()
        self.header.add_headers(self._content.header_lines())

    def is_complete(self):
        """ Has the entire input stream been seen?
        :return: True or False
        """
        if self.header.finished is False:
            return False
        elif not self.header.needs_content:
            return True
        return self._content.finished

    def add_content(self, cntnt):
        self._content.add_content(cntnt)



from _socket import SHUT_RD, SHUT_RDWR
import logging
import socket
import threading
import select
from atavism.http11.objects import HttpRequest


class HttpServerError(Exception):
    pass


class HttpServer(object):
    """ Class that serves an HLSVideo via HTTP.
    """

    def __init__(self, host=None, port=80, handler=None):
        self.socket = None
        self.backlog = 5
        self.running = False
        self.accept_thread = None
        self.connections = []
        self.logger = logging.getLogger()
        self.host = host
        self.port = port

        if not hasattr(self, 'handler'):
            self.handler = handler

    def __del__(self):
        if self.socket is not None:
            self.socket.close()

    @property
    def is_valid(self):
        return self.socket is not None

    def make_socket(self):
        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            sock.setblocking(0)
            sock.settimeout(5)
            sock.bind((self.host, self.port))
            sock.listen(self.backlog)
            self.socket = sock
        except socket.error:
            raise HttpServerError("Unable to open a socket for {}:{}".format(self.host, self.port))

    def start(self):
        self.logger.info("HttpServer starting up")
        if self.socket is None:
            self.make_socket()
            if self.socket is None:
                raise HttpServerError("Failed to start as no socket was created.")

        self.running = True
        self.accept_thread = threading.Thread(target=self._accept_loop)
        self.accept_thread.start()

    def stop(self):
#        if self.socket is not None:
#            self.socket.shutdown(SHUT_RD)
#            self.socket.close()
#            self.socket = None

        for c in self.connections:
            c.stop()

        if self.running:
            self.running = False
            try:
                self.accept_thread.join()
            except KeyboardInterrupt as e:
                pass

        if self.socket is not None:
            self.socket.shutdown(SHUT_RD)
            self.socket.close()
            self.socket = None


    def join(self):
        self.accept_thread.join(1.0)

    def _accept_loop(self):
        self.make_socket()
        self.logger.info("Starting accept loop for {}:{}".format(self.host, self.port))
        while self.running:
            r, w, e = select.select([self.socket], [], [self.socket], 5.0)
            if len(e) > 0:
                break
            if len(r) == 0:
                continue

            try:
                ns = self.socket.accept()
                conn = HttpConnection(self, *ns)
                self.connections.append(conn)
                self.logger.info("Accepted a connection from %s.", ns)
            except socket.timeout as e:
                continue
            except (OSError, socket.error, AttributeError):
                if self.running:
                    self.logger.warning("Socket error in accept loop :-(")
                break

        self.running = False


class HttpConnection(object):
    def __init__(self, parent, sock, address):
        self.parent = parent
        self.socket = sock
        self.address = address
        self.logger = logging.getLogger()

        self.inp = b''
        self.running = True
        self.responses = []

        self.thread = threading.Thread(target=self.main_loop)
        self.thread.name = 'HttpConnection'
        self.thread.daemon = True
        self.thread.start()
        parent.connections.append(self)

    def main_loop(self):
        #todo add timeout checking...
        request = None
        self.logger.info("HttpConnection main loop started.")
        while self.running:
            ws = [self.socket] if len(self.responses) > 0 else []

            try:
                r, w, e = select.select([self.socket], ws, [self.socket], 5.0)
            except:
                break

            if len(e) > 0:
                break

            resp = self.responses[0] if len(self.responses) > 0 else None
            if len(r) > 0:
                try:
                    data = self.socket.recv(2048)
                except socket.error as e:
                    self.logger.warning("Socket error: %s", e)
                    break

                self.logger.debug("Read %d bytes from accepted socket", len(data))
                if len(data) == 0:
                    if resp is None or not resp.is_keepalive:
                        self.logger.debug("Zero byte read, keepalive not asked for, closing socket...")
                        break

                self.inp += data
                if request is None:
                    request = HttpRequest()

                read = request.read_content(self.inp)
                self.inp = self.inp[read:]

                if request.header.finished:
                    self.logger.debug(request.header)

                if request.is_complete():
                    resp = self.parent.handler(request)
                    resp.complete()
                    self.responses.append(resp)
                    request = None

            if len(w) > 0 and resp is not None:
                # only process one response each pass...
                next = resp.next_output()
                self.logger.debug("Sending %s bytes", len(next))
                try:
                    self.socket.send(next)
                except socket.error as e:
                    if self.running:
                        self.logger.warning("Unable to send via socket. Closing it. %s", e)
                    break
                if resp.send_complete():
                    if not resp.is_keepalive:
                        break
                    self.responses.pop()

        self.socket.close()
        self.running = False
        self.parent.connections.remove(self)

    def stop(self):
        self.running = False
        try:
            self.socket.shutdown(SHUT_RDWR)
        except socket.error:
            pass



import gzip
import json
from lxml import etree
import mimetypes
import os
import zlib

try:
    from cStringIO import StringIO as GzipIO
except ImportError:
    from io import BytesIO as GzipIO

try:
    from plistlib import loads as plist_loads
except ImportError:
    from plistlib import readPlistFromString as plist_loads


class Content(object):
    """ Class to manage content for an HTTP transaction (i.e. a Request or a Response).
        The class can accept content from a network stream via read_content() or
        directly using add_content.
        The class maintains enough information about it's contents to allow HTTP headers
        to be generated.
        As each instance can contain only one type of data, if the data is compressed or
        manipulated another instance will be created with the new data, and the 'next' field
        will point to it.
        All functions that access/manipulate data handle this case, so the actual content of
        an instance should never be accessed directly. Calling the appropriate function on the
        root object in a chain will return the expected data.
    """
    CRLF = b'\r\n'
    RANGE_BOUNDARY = 'One_At_A_Time_Please'
    MAX_SEND = 2048

    def __init__(self, data=None, content_sz=None, content_type=None, charset=None):
        self._buffer = b''
        self._next = None

        self.compression = False
        self.is_compressed = False
        self.finished = False
        self.send_complete = False
        self.content_type = content_type
        self.content_sz = content_sz
        self.charset = charset
        self.send_position = 0

        if self.content_sz == 0:
            self.finished = True

        if self.content_sz is not None and data is not None:
            self.read_content(data)
        elif data is not None:
            self.add_content(data)

    def __len__(self):
        """ Return the actual length of the content. When subclassed, this function may be overwritten,
            e.g. FileContent returns the filesize if the file exists. For this reason, within the class
            use len(self) rather than len(self._buffer) unless the intent is carefully considered.
        :return: Number of bytes available.
        """
        if self._next is not None:
            return len(self._next)
        return len(self._buffer)

    def __getitem__(self, item):
        if self._next is not None:
            return self._next.__getitem__(item)
        return self._buffer.__getitem__(item)

    def reset(self):
        self.compression = None
        self.content_type = None
        self.charset = None
        self.content_sz = None
        self.send_position = 0
        self._buffer = b''
        self._next = None
        self.finished = False
        self.is_compressed = False

    def set_compression(self, method):
        if self._next is not None:
            self._next.set_compression(method)
        else:
            self.compression = method

    def compress(self):
        if self._next is not None:
            self._next.compress()
        if self.is_compressed:
            return
        elif self.compression == 'gzip':
            zbuf = GzipIO()
            zfile = gzip.GzipFile(mode='wb',  fileobj=zbuf, compresslevel=9)
            zfile.write(self._buffer)
            zfile.close()
            ct = self._add_next(zbuf.getvalue())
            ct.is_compressed = True
            ct.compression = 'gzip'
        elif self.compression == 'deflate':
            ct = self._add_next(zlib.compress(self._buffer))
            ct.is_compressed = True
            ct.compression = 'deflate'

    def decompress(self):
        if self._next is not None:
            self._next.decompress()
        elif self.compression == 'gzip':
            self._add_next(gzip.GzipFile('', 'r', 0, GzipIO(self._buffer)).read())
        elif self.compression == 'deflate':
            self._add_next(zlib.decompress(self._buffer))

    ### Adding content.
    def read_content(self, cntnt):
        """ Read content from a stream and return how many bytes have been read.
            * Chunked Transfer Encoding is handled in this function as the final content length is unknown.
            * If the final content length is known, this is honoured and content will be capped at that length.
            * If not chunked and no content length is set, we simply accept all data provided.
            As this is often used with streams of data, return the number of bytes from the provided data have
            been read.
        :param cntnt: The content to be added.
        :return: Number of bytes added.
        """
        if self._next is not None:
            return self._next.read_content(cntnt)
        consumed = 0
        if self.finished:
            return consumed

        if self.content_sz == 'chunked':
            if self.CRLF not in cntnt:
                return 0
            pos = 0
            while True:
                if self.CRLF not in cntnt[pos:]:
                    return pos

                chunk_len, ignored = cntnt[pos:].split(self.CRLF, 1)
                cr = len(chunk_len) + 2
                chunk_len = int(chunk_len, 16)

                if pos + cr + chunk_len + 2 > len(cntnt):
                    break
                else:
                    pos += cr
                    if chunk_len > 0:
                        self._buffer += cntnt[pos:pos + chunk_len]
                    pos += chunk_len + 2
                    consumed = pos

                if chunk_len == 0:
                    self.finished = True
                    consumed = pos
                    break
        elif self.content_sz is None:
            self._buffer += cntnt
            consumed = len(cntnt)
        else:
            consumed = min(self.content_sz - len(self), len(cntnt))
            self._buffer += cntnt[:consumed]
            if self.content_sz == len(self):
                self.finished = True

        if self.finished:
            self.decompress()
        return consumed

    def add_content(self, cntnt):
        """ Add content to the buffer. If the data is from a network stream, read_content() should be used instead.
        :param data: The data to be added.
        :return: None.
        """
        if self._next is not None:
            return self._next.add_content(cntnt)
        self._buffer += cntnt if isinstance(cntnt, bytes) else cntnt.encode()

    @property
    def content(self):
        """ Get the full, raw content.
        :return: The objects content.
        """
        if self._next is not None:
            return self._next.content
        return self._buffer

    def decoded_content(self):
        if self._next is not None:
            return self._next.decoded_content()

        self.check_content_type()
        if self.content_type is None or self.content_type in ('text/plain', 'text/html'):
            if self.charset is not None:
                return self._buffer.decode(self.charset)
            return self._buffer.decode()

        try:
            if self.content_type in ['text/x-apple-plist+xml', 'application/x-apple-binary-plist']:
                return plist_loads(self._buffer)
            elif self.content_type == 'text/parameters':
                dd = {}
                for line in self._buffer.split(b'\n'):
                    if line == b'':
                        continue
                    k, v = line.split(b':', 1)
                    dd[k] = v.strip()
                return dd
            elif self.content_type == 'application/json':
                return json.loads(self._buffer.decode())
            elif self.content_type == 'application/xml':
                return etree.fromstring(self._buffer.decode())
            elif self.content_type == 'multipart/byteranges':
                boundary = self.charset.split('=', 1)[1]
                start = self._buffer.find(b'--')
                if start == -1:
                    return []
                parts = self._buffer[start:].strip().split("--{}".format(boundary).encode())
                return [self.parse_range_multipart(p) for p in parts if len(p) > 2]
        except:
            return self._buffer

    def parse_range_multipart(self, p):
        obj = {}
        if b'\r\n\r\n' not in p:
            return {}
        hdrs, obj['content'] = p.strip().split(b'\r\n\r\n')
        for hdr in hdrs.split(self.CRLF):
            if b':' in hdr:
                k, v = hdr.decode().split(':')
                obj[k.strip()] = v.strip()
        if 'Content-Type' in obj and obj['Content-Type'] in 'text/html':
            obj['content'] = obj['content'].decode()
        return obj

    def check_content_type(self):
        if self.content_type is not None and ';' in self.content_type:
            self.content_type, self.charset = [p.strip() for p in self.content_type.split(';')]
            self.charset = self.charset.replace('charset=', '').strip()

    def header_lines(self):
        if self._next is not None:
            return self._next.header_lines()
        rv = {}

        if self.content_type is not None:
            if self.charset is None:
                rv['Content-Type'] = self.content_type
            else:
                rv['Content-Type'] = '{}; charset={}'.format(self.content_type, self.charset)
        if self.content_sz == 'chunked':
            rv["Transfer-Encoding"] = "chunked"
        elif len(self):
            rv["Content-Length"] = "{}".format(len(self))
        if self.compression:
            rv["Content-Encoding"] = self.compression
            rv["Vary"] = 'Content-Encoding'
        return rv

    def next(self, pkt_len):
        if self._next is not None:
            return self._next.next(pkt_len)
        if len(self) == 0:
            self.send_complete = True
            return b''
        avail = min(self.MAX_SEND - pkt_len, len(self) - self.send_position)
        if avail == 0:
            self.send_complete = True
            return b''
        if self.content_sz == 'chunked':
            avail -= 8
        rv = self[self.send_position: self.send_position + avail]
        self.send_position += len(rv)
        if self.content_sz == 'chunked':
            return "{:4X}\r\n{}\r\n".format(len(rv), rv)
        return rv

    def create_ranged_output(self, ranges):
        """ If we have been asked for byte ranges, then we create them in a new Content object with
            the appropriate formatting. We need to use the current content for the ranges, but as that
            content could have been manipulated already (or be a file object), we need to access it
            correctly.
        :param ranges: List of the ranges to be processed.
        :return: None.
        """
        hdrs = {}
        if self._next is not None:
            return self._next.create_ranged_output(ranges)

        if len(ranges) == 0:
            return

        self.check_content_type()
        ct = self.content_type if len(ranges) == 1 else 'multipart/byteranges; boundary={}'.format(self.RANGE_BOUNDARY)
        cntnt = Content(content_type=ct, charset=self.charset)
        content_len = len(self)

        if len(ranges) > 1:
            for r in ranges:
                cntnt.add_content("--{}\r\nContent-Type: {}\r\nContent-Range: bytes={}\r\n\r\n".format(
                    self.RANGE_BOUNDARY, self.content_type, r.absolute_range(content_len)))
                start, end = r.absolutes(content_len)
                cntnt.add_content(self[start:end + 1])
            cntnt.add_content("--{}--\r\n".format(self.RANGE_BOUNDARY))
        else:
            start, end = ranges[0].absolutes(content_len)
            cntnt.add_content(self[start: end + 1])
            hdrs['Content-Range'] = 'bytes {}'.format(ranges[0].absolute_range(content_len))
        self._next = cntnt
        return hdrs

    def _add_next(self, data=None, content_type=None):
        self.check_content_type()
        ct = Content(data=data,
                     content_type=content_type or self.content_type,
                     charset=self.charset)
        self._next = ct
        return ct

class FileContent(Content):
    def __init__(self, filename):
        Content.__init__(self, content_sz=os.path.getsize(filename))
        self.filename = filename
        self.file_handle = None
        self.exists = os.path.exists(filename)
        if not self.exists:
            return
        self.content_type, ignored = mimetypes.guess_type(filename)

    def __del__(self):
        if self.file_handle is not None:
            self.file_handle.close()

    def __len__(self):
        if self._next is not None:
            return len(self._next)
        if self.exists:
            return self.content_sz
        return len(self._buffer)

    def _open(self):
        """ Open the file for reading.
        :return: None
        """
        if self.file_handle is not None:
            return
        self.file_handle = open(self.filename, "rb")

    def _close(self):
        """ Close the open file handle.
        :return: None
        """
        if self.file_handle is None:
            return
        self.file_handle.close()
        self.file_handle = None

    def __getitem__(self, item):
        if isinstance(item, slice):
            self._open()
            self.file_handle.seek(item.start)
            if self.file_handle.tell() == item.start:
                return self.file_handle.read(item.stop - item.start)

        elif isinstance(item, int):
            self._open()
            self.file_handle.seek(item)
            if self.file_handle.tell() == item:
                return self.file_handle.read(1)

    def write(self):
        if self._next is not None:
            if len(self._next) == 0:
                return
        if len(self._buffer) == 0:
            return
        self._close()
        with open(self.filename, "wb") as fh:
            fh.write(self._buffer)



""" HTTP relies on Requests and Responses. The classes contained in this module are
    able to be created in either a client or server environment. They are named for
    their content and intent.
"""
try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse

from atavism.http11.base import BaseHttp


class HttpRequest(BaseHttp):
    """ An HTTP Request.
    """
    def __init__(self, inp=None, method=None, path=None):
        BaseHttp.__init__(self)
        if inp is not None:
            self.read_content(inp)
        self.method = method or 'GET'
        if path is not None:
            self.path = path if not isinstance(path, bytes) else path.decode()

    def read_content(self, data):
        if data is None or len(data) == 0:
            return 0
        hdr = self.header.finished
        rv = BaseHttp.read_content(self, data)
        if hdr is False and self.header.finished:
            self.method, self.path, self.http = self.header.status_line.split(' ')
        return rv

    def reset(self):
        cnt = self._content
        while cnt._next:
            cnt = cnt._next
        cnt.send_position = 0
        self.headers_sent = False

    def complete(self):
        if len(self.ranges) > 0:
            self.add_header('Range', 'bytes={}'.format(','.join([r.header() for r in self.ranges])))
        self.header.status_line = '{} {} {}'.format(self.method.upper(), self.path, self.http)
#        if len(self._content) > 512:
#            self._content.content_sz = 'chunked'
        self._complete()

    def make_response(self):
        resp = HttpResponse()

        resp.add_header('Connection', 'keep-alive' if self.is_keepalive else 'close')
        resp.ranges = self.ranges

        if self.method.upper() == b'HEAD':
            resp.headers_only = True

        ce = self.get('accept-encoding')
        if ce is not None and 'gzip' in ce:
            # todo - handle this properly
            resp.set_compression('gzip')
        return resp


class HttpResponse(BaseHttp):
    """ Class that represents a response from an HTTP server.
    """
    #todo - expand list
    STATUS_MSG = {
        200: 'OK',
        206: 'Partial Content',
        301: 'Moved permanently',
        401: 'Unathorised',
        402: 'Payment required',
        403: 'Forbidden',
        404: 'Not found',
        405: 'Method not allowed',
        416: 'Requested range not satisfiable'
    }

    def __init__(self, inp=None, code=None):
        BaseHttp.__init__(self)
        self.msg = 'OK'
        self.code = code or 200
        if inp is not None:
            self.read_content(inp)

    def read_content(self, data):
        if data is None or len(data) == 0:
            return 0
        hdr = self.header.finished
        rv = BaseHttp.read_content(self, data)
        if hdr is False and self.header.finished:
            self.http, self.code, self.msg = self.header.status_line.split(' ', 2)
            self.code = int(self.code)
        return rv

    def status_msg(self):
        return self.STATUS_MSG.get(self.code, "Unknown status! {}".format(self.code))

    def set_code(self, code):
        """ Set the ode to use for the response and set
        :param code: The numeric error code to respond with.
        :return: None.
        """
        self.code = code
        if code >= 400:
            self.ranges = []
        elif code == 206 and len(self.ranges) == 0:
            self.code = 200

    def mark_complete(self):
        self._content.finished = True
        self._content.decompress()

    def complete(self):
        if len(self.ranges) > 0:
            self.check_ranges()
            if self.code == 200:
                self.code = 206

        hdrs = self._content.create_ranged_output(self.ranges)
        self.header.add_headers(hdrs)
        self.header.status_line = 'HTTP/1.1 {} {}'.format(self.code, self.status_msg())
        self._complete()

    def check_ranges(self):
        for r in self.ranges:
            st, end = r.absolutes(len(self))
            if 0 < st >= len(self) or st > end >= len(self):
                self.set_code(416)
                break



""" A multithreaded, keep-alive HTTP/1.1 client object that can handle multiple
    connections.
"""
import socket
import select

try:
    from urllib.parse import urlencode, quote
except ImportError:
    from urllib import urlencode, quote

from atavism import __version__
from atavism.http11.objects import HttpRequest, HttpResponse
from atavism.http11.cookies import CookieJar


class HttpClientError(Exception):
    pass


class HttpClient(object):
    """ Http Client class. Implements an HTTP 1.1 client which uses keepalive by default.
        Requests can always be made, but may block if another request is being processed.
    """
    TIMEOUT = 5.0

    def __init__(self, host, port=80):
        self.host = host
        self.port = port
        self.socket = None

        self.cookies = CookieJar()
        self.user_agent = 'atavism/{}'.format(__version__)

        self.timeout = self.TIMEOUT
        self._buffer = b''

        if isinstance(self.host, bytes):
            self.host = self.host.decode()

    def __del__(self):
        self._close_socket()

    def host_str(self):
        if self.port == 80:
            return self.host
        return "{}:{}".format(self.host, self.port)

#    def add_cookie(self, key, value, path='/'):
#        self.cookies.set_cookie(key, value, path)

    def verify(self):
        """ Verify that the client is able to establish a connection to the server.
            NB This will not start the listener thread.
        :return: True or False
        """
        self._make_socket()
        if self.socket is not None:
            return True

    def simple_request(self, uri=None, qry=None):
        """ Really, really simple GET request. This will not follow redirects and for anything but a
            200 response will return None.
        :param uri: The server URI to GET
        :param qry: Optional query string to append.
        :return: The content of the response or None.
        """
        resp = self.request(uri, qry)
        return resp.decoded_content()

    def download_file(self, uri, filename):
        resp = self.request(uri)
        if resp is None:
            return False
        if resp.code != 200:
            raise HttpClientError("Unable to download file. Error code {}".format(resp.code))
        with open(filename, 'wb') as fh:
            fh.write(resp.content)
        return True

    def request(self, uri, qry=None):
        return self._make_send_request('GET', uri, qry=qry)

    def post_data(self, uri, qry=None, data=None, ct=None):
        """ Submit a POST request with the supplied data. """
        if ct is None and data is not None and len(data) > 0:
            ct = 'application/x-www-form-urlencoded'
        hdrs = {'Content-Type': ct}
        if data is not None:
            if ct == 'text/parameters':
                data = "\r\n".join("{}: {}".format(k, data[k]) for k in data) + "\r\n"
            elif isinstance(data, dict):
                data = urlencode(data).encode()
        return self._make_send_request('POST', uri, qry=qry, data=data, hdrs=hdrs)

    def _make_url(self, path, query=None):
        if path is None or len(path) == 0:
            path = b'/'
        else:
            path = quote(path)

        if query is not None:
            if isinstance(query, dict):
                query = urlencode(query)
            if '?' in path:
                return path + query
            return path + '?' + query
        return path

    def create_request(self, method, uri='/', qry=None, hdrs=None):
        req = HttpRequest(method=method, path=self._make_url(uri, qry))
        req.add_headers(hdrs or {})
        cookies = self.cookies.get_cookies(uri)
        if cookies is not None:
            req.add_header('Cookie', cookies)
        return req

    def _make_send_request(self, method, uri='/', qry=None, data=None, hdrs=None):
        req = self.create_request(method, uri, qry=qry, hdrs=hdrs)
        if data is not None:
            req.add_content(data)
        return self.send_request(req)

    def send_request(self, request):
        self._make_socket()
        if self.socket is None:
            return None

        request.add_headers({'Host': self.host_str(),
                             'Accept-Encoding': 'identity, gzip'})
        if self.user_agent:
            request.add_header('User-Agent', self.user_agent)
        request.complete()

        response = self._process_request(request)
        if response is None:
            raise HttpClientError("No response received from remote server.")
        return response

    def _make_socket(self):
        """ Create a new connected socket.
        :raise HttpClientError:
        """
        if self.socket is None:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
            sock.settimeout(5)
            try:
                sock.connect((str(self.host), self.port))
                sock.settimeout(30)
                self.socket = sock
            except socket.error:
                raise HttpClientError("Attempt to connect to '{}' on port {} failed".format(self.host, self.port))
            except socket.timeout:
                raise HttpClientError("Attempt to connect to '{}' on port {} timed out".format(self.host, self.port))
            except socket.gaierror:
                raise HttpClientError("Unable to resolve host '{}'".format(self.host))

    def _close_socket(self):
        if self.socket is not None:
            self.socket.close()
            self.socket = None
        self._buffer = b''

    def _process_request(self, request):
        """ Process a single request.
        :return: None or the HttpResponse
        """
        self._make_socket()
        # Send the entire request...
        while not request.send_complete():
            data = request.next_output()
#            print("send({})".format(data))
            if len(data) == 0:
                break

            try:
                r, w, e = select.select([], [self.socket], [self.socket], self.timeout)
            except socket.error:
                print("Select failed for socket - trying to reestablish...")
                self._close_socket()
                request.reset()
                return self._process_request(request)

            if len(e) > 0:
                raise HttpClientError("Socket reported an error.")
            elif len(w) == 0:
                raise HttpClientError("Socket timed out for write operations. Unable to send request.")
            n = self.socket.send(data)
            if n == 0:
                print("Unable to send data")
                break

        if not request.send_complete():
            raise HttpClientError("Unable to send the request.")

        # Receive the whole response...
        response = HttpResponse(self._buffer)
        while not response.is_complete():
            r, w, e = select.select([self.socket], [], [self.socket], self.timeout)
            if len(e):
                self._close_socket()
                break

            if len(r):
                data = self.socket.recv(2048)
                if len(data) == 0:
                    response.mark_complete()
                    break
                self._buffer += data
#                print(self._buffer)
                r = response.read_content(self._buffer)
                self._buffer = self._buffer[r:]

            if response.is_complete():
                break

        if response.is_complete():
            self.cookies.check_cookies(response)
            if not response.is_keepalive:
                self._close_socket()
            return response
        return None



from setuptools import setup, find_packages
from os import path
from atavism import __version__
import io


here = path.abspath(path.dirname(__file__))

# Get the long description from the relevant file
with io.open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


setup(
    name='atavism',
    version=__version__,
    description='Module to play video to an AppleTV',
    long_description=long_description,
    url='https://github.com/zathras777/atavism',
    author='david reid',
    author_email='zathrasorama@gmail.com',
    license='Unlicense',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Multimedia :: Video :: Display',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
    ],
    keywords='appletv video ffmpeg streaming',
    packages=find_packages(exclude=['tests']),
    test_suite='tests',
    install_requires=[
        'ipaddress'
    ],
    entry_points={
        'console_scripts': ['atavism=atavism.command_line:main']
    },
    download_url='https://github.com/zathras777/atavism/tarball/0.2',
)



# -*- coding: utf-8 -*-
#

# RFC 1035 - DNS
# RFC 6762 - Multicast DNS
# RFC 6763 - DNS Service Discovery

import ipaddress
import logging
import random
import time
import socket
import select
import struct
import sys

PY3 = True if sys.version_info[0] == 3 else False

QTYPE_A = 1
QTYPE_NS = 2
QTYPE_PTR = 12
QTYPE_TXT = 16
QTYPE_AAAA = 28
QTYPE_SRV = 33
QTYPE_OPT = 41
QTYPE_NSEC = 47
QTYPE_ALL = 255

QCLASS_IN = 1
QCLASS_NONE = 254
QCLASS_ANY = 255

FLAGS_RCODE = (1 << 0) | (1 << 1) | (1 << 2) | (1 << 3)
FLAGS_Z = (1 << 4) | (1 << 5) | (1 << 6)
FLAGS_RA = 1 << 7
FLAGS_RD = 1 << 8
FLAGS_TC = 1 << 9
FLAGS_AA = 1 << 10  # authoritative
FLAGS_OPCODE = (1 << 11) | (1 << 12) | (1 << 13) | (1 << 14)
FLAGS_QR = 1 << 15  # response

OPCODE_QUERY = 1
OPCODE_STATUS = 2
OPCODE_NOTIFY = 4
OPCODE_NOTIFY = 5


class PacketError(Exception):
    pass


class Packet(object):
    """ Class to represnt a packet received from or being sent via the network.
        Contains all the data and functions to operate on the data.
        The data is stored as bytes internally but returned as str objects normally.
    """
    def __init__(self, data=None):
        self.txt_offsets = {}
        self.data = b''
        self.add_data(data)

    def reset(self):
        self.data = b''

    def add_data(self, data):
        if data is None or len(data) == 0:
            return
        if isinstance(data, str):
            self.data += data.encode() if PY3 else data
        elif isinstance(data, bytes):
            self.data += data
        else:
            raise PacketError("Trying to add data of type '{}' to the buffer failed.".format(type(data)))

    def __len__(self):
        return len(self.data)

    def read_name(self, pos=0):
        """ Read an encoded name from the data. The name will consist of a series of pieces,
            each of which may be a string or an offset to the string.
            If a name cannot be read, a PacketError exception will be raised.
        :param pos: The position within the data to start reading the name.
        :return: The number of bytes of the stream read, the namename.
        """
        bytes, name, eos = self._read_name(pos)
        if not eos:
            raise PacketError("Unable to read a name from position {}".format(pos))
        return bytes, name

    def _read_name(self, pos):
        """ Internal function to get a name, or portion of a name from the data.
            May be called recursively.
        """
        start = pos
        eos = False
        parts = []
        while True:
            len_or_offset = ord(self.data[pos]) if not PY3 else self.data[pos]
            if len_or_offset == 0:
                pos += 1
                eos = True
                break

            if len_or_offset & 0xc0 == 0:
                pos += 1
                # simple length of string that follows
                if len_or_offset == 0:
                    eos = True
                else:
                    utf8 = self.read_utf8(pos, len_or_offset)
                    pos += len_or_offset
                    parts.append(utf8)

            elif len_or_offset & 0xc0 == 0xC0:
                # offset to string
                offset = struct.unpack("!H", self.data[pos:pos+2])[0]
                offset ^= 0xc000
                pos += 2
                ignored, utf8, eos = self._read_name(offset)
                parts.append(utf8)
                if eos:
                    break

            else:
                raise PacketError("Bad domain name at {0:X}".format(self.data[pos]))
        return pos - start, '.'.join(parts), eos

    def write_name(self, name):
        """ Write an encoded name into the data stream.
        :param name: The name to be encoded.
        :returns: The number of bytes added to the data stream.
        """
        # Do we have the entire name available?
        if name in self.txt_offsets:
            offs = self.txt_offsets[name]
            offs |= (1 << 15)
            offs |= (1 << 14)
            self.pack("!H", offs)
            return 2
        # Add the
        start = len(self.data)
        parts = name.split('.')
        null_needed = True
        for i in range(len(parts)):
            pstr = ".".join(parts[i:])
            if pstr in self.txt_offsets:
                offs = self.txt_offsets[pstr]
                offs |= (1 << 15)
                offs |= (1 << 14)
                self.pack("!H", offs)
                null_needed = False
                break
            else:
                self.txt_offsets[pstr] = len(self.data)
                utf8_string = parts[i].encode('utf-8')
                self.pack('!B', len(utf8_string))
                self.data += utf8_string
        if null_needed:
            self.pack("!b", 0)
        return len(self.data) - start

    def read_utf8(self, pos, len):
        """ Attempt to read a utf-8 string from the stored data, blen bytes starting at
            bpos offset into the stream. Return the number of bytes of the stream read
            to the end of the string and the string.
        """
        return self.data[pos: pos + len].decode('utf-8')

    def ipaddress(self, version, pos, nbytes):
        if version == 4:
            return ipaddress.IPv4Address(self.data[pos:pos+nbytes])
        elif version == 6:
            return ipaddress.IPv6Address(self.data[pos:pos+nbytes])
        return None

    def pack(self, fmt, *args, **kwargs):
        """ Attempt to write packed data into the data. This is simply appended to the end of the
            data.
        :param fmt: The format of the data to be written.
        :param data: Data to be written.
        :return: Number of bytes written.
        """
        blen = struct.calcsize(fmt)
        if PY3:
            data = [a.encode() if isinstance(a, str) else a for a in args]
            pdata = struct.pack(fmt, *data)
        else:
            pdata = struct.pack(fmt, *args)

        if 'pos' in kwargs:
            pos = kwargs['pos']
            if 0 < pos > len(self.data):
                raise PacketError("Attempt to pack data outside of available data (@ {} but data is {} bytes)".format(pos, len(self.data)))
            self.data = self.data[:pos] + pdata + self.data[pos+blen:]
        else:
            self.data += pdata
        return blen

    def unpack(self, pos, fmt):
        """ Attempt to unpack a binary format from the data. If this returns a single item then that
            item is returned, but if more than one value is unpacked a list of those values will be
            returned.
            i.e. unpack(..., "b") will return a byte value, unpack(..., '"bb") will return a list of 2 byte
                 values.
            NB If the position specified is outside the size of the buffer, a PacketError exception is raised.
        :param pos: Position within buffer to start
        :param fmt: Format to be unpacked
        :return: Length of bytes read and value(s).
        """
        blen = struct.calcsize(fmt)
        if pos < 0 or pos + blen > len(self.data):
            raise PacketError("Invalid position specified for unpack.")
        parts = [n.decode() if isinstance(n, bytes) else n for n in struct.unpack(fmt, self.data[pos:pos + blen])]
        if len(parts) == 1:
            return blen, parts[0]
        return blen, parts


class MDNSResponse(object):
    TYPE_NAMES = {
        1: 'A',
        12: 'PTR',
        16: 'TXT',
        28: 'AAAA',
        33: 'SRV',
    }

    def __init__(self, data=None):
        self.packet = Packet(data)

        self.pos = 0
        self.questions = []
        self.answers = []
        self.nameservers = []
        self.additional = []

        self.pos, parts = self.packet.unpack(0, "!HHHHHH")
        self.id, self.flags, self.qdcount, self.ancount, self.nscount, self.arcount = parts

        # Is it a response? Do we have answers?
        if self.flags & FLAGS_QR == 0 or self.ancount == 0 or self.flags & FLAGS_RCODE != 0:
            return
        self.parse_records()

    @property
    def is_valid(self):
        return len(self.answers) != 0

    @property
    def has_error(self):
        return self.flags & FLAGS_RCODE != 0

    def is_applicable(self, qry):
        qnames = [q['qname'] for q in qry.questions]
        for a in self.answers:
            if a['qname'] in qnames:
                return True
            if ".".join(a['qname'].split('.')[1:]) in qnames:
                return True
        return False

    @property
    def is_authoritative(self):
        return self.flags & FLAGS_AA != 0

    def parse_records(self):
        for q in range(self.qdcount):
            n, qname = self.packet.read_name(self.pos)
            qu = {'qname': qname}
            self.pos += n
            n, (qu['qtype'], qu['qclass']) = self.packet.unpack(self.pos, "!HH")
            self.pos += n
            self.questions.append(qu)

        self._parse_section(self.ancount, self.answers)
        self._parse_section(self.nscount, self.nameservers)
        self._parse_section(self.arcount, self.additional)

    def _parse_section(self, num, store):
        for a in range(num):
            n, qname = self.packet.read_name(self.pos)
            self.pos += n

            n, (typ, cls, ttl, rdlength) = self.packet.unpack(self.pos, "!HHiH")
            rec = {'qname': qname, 'qtype': typ, 'qclass': cls, 'ttl': ttl}
            self.pos += n

            if typ == QTYPE_SRV:
                n, (priority, weight, port) = self.packet.unpack(self.pos, "!HHH")
                self.pos += n
                n, name = self.packet.read_name(self.pos)
                rec['SRV'] = {'port': port,
                              'name': name,
                              'priority': priority,
                              'weight': weight}
                self.pos += n
            elif typ == QTYPE_A:
                rec['A'] = self.packet.ipaddress(4, self.pos, rdlength)
                self.pos += rdlength
            elif typ == QTYPE_AAAA:
                rec['AAAA'] = self.packet.ipaddress(6, self.pos, rdlength)
                self.pos += rdlength
            elif typ == QTYPE_PTR:
                n, rec['PTR'] = self.packet.read_name(self.pos)
                self.pos += n
            elif typ == QTYPE_TXT:
                rec['TXT'] = self.packet.data[self.pos: self.pos + rdlength]
                self.pos += rdlength
            elif typ == QTYPE_OPT:
                rec['OPT'] = self.packet.data[self.pos: self.pos + rdlength]
                self.pos += rdlength
            else:
                self.pos += rdlength
            store.append(rec)


class MDNSQuery(object):
    MAX_ANSWERS = 24

    def __init__(self, questions=None, answers=None):
        self.questions = questions or []
        self.answers = answers or []
        self.packet = Packet()
        self.pkt_id = random.getrandbits(16)
        self.flags = 0

    def __len__(self):
        return len(self.questions)

    def add_question(self, qname, qtype=QTYPE_ALL, qclass=QCLASS_IN):
        self.questions.append({'qname': qname, 'qtype': qtype, 'qclass': qclass})

    def add_answer(self, qname, ptr, qtype, qclass=QCLASS_IN, ttl=0):
        self.answers.append({'qname': qname, 'ptr': ptr, 'qtype': qtype, 'qclass': qclass, 'ttl': ttl})

    def packet_data(self):
        """ Create the packets we will send. There could be more than 1 packet if we have a
            large number of answers to include. Answers are included to reduce network traffic.
            As the query is created, then used for each sending attempt, we increment the pkt_id
            each time we create the packets. All packets
        :return: A list of packets.
        """
        idx = 0
        packets = []
        while idx <= len(self.answers):
            flags = self.flags
            pkt = Packet()

            if len(self.answers) - idx >= self.MAX_ANSWERS:
                flags |= FLAGS_TC

            # We only include the question on the initial packet...
            qc = len(self.questions) if idx == 0 else 0
            ac = min(self.MAX_ANSWERS, len(self.answers) - idx)

            if qc + ac == 0:
                break

            pkt.pack("!HHHHHH", self.pkt_id, flags, qc, ac, 0, 0)

            for q in self.questions[:qc]:
#                print("  Q: {}".format(q))
                pkt.write_name(q['qname'])
                pkt.pack("!HH", q['qtype'], q['qclass'])

            for a in self.answers[idx: idx + ac]:
#                print("  A: {}".format(a))
                pkt.write_name(a['qname'])
                pkt.pack("!HHI", a['qtype'], a['qclass'], a['ttl'])
                rdpos = len(pkt)
                pkt.data += b'  '
                ptrlen = pkt.write_name(a['ptr'])
                pkt.pack("!H", ptrlen, pos=rdpos)

            packets.append(pkt.data)
            if ac == 0:
                break
            idx += ac
        return packets


class MDNSServiceDiscoveryError(Exception):
    pass


class MDNSServiceDiscovery(object):
    IP4_MULTICAST = ipaddress.IPv4Address(u'224.0.0.251')
    IP6_MULTICAST = ipaddress.IPv6Address(u'FF02::FB')
    MULTICAST_PORT = 5353

    def __init__(self, *args, **kwargs):
        self.ip_version = 4
        self.ttl = 2
        self.timeout = 10
        self.devices = {}
        self.qtype = kwargs.get('qtype', QTYPE_ALL)
        self.interface = self.find_interfaces()
        self.query = MDNSQuery()

        self.logger = logging.getLogger(__name__)
        self.logger.setLevel(logging.getLogger().level)

        for qname in args:
            self.query.add_question(qname.strip(), self.qtype)

    def find_interfaces(self):
        """ Find the local interface(s) that we will send via.
            Presently I have no way of finding a local IPv6 interface, so just use
            IPv4.
            An MDNSServiceDiscoveryError will be raised if the local interface can't be found.
        :return: The interface to be used.
        """
        try:
            x = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            x.connect(('1.2.3.4', 56))
            interface = x.getsockname()[0]
            x.close()
        except socket.error as e:
            raise MDNSServiceDiscoveryError("Unable to find local interface.")
        return interface

    def find_devices(self):
        now = time.time()
        nxt = now
        last = now + self.timeout
        delay = 1

        sock = self.make_socket()
        while last > now:
            now = time.time()

            if now >= nxt:
                for data in self.query.packet_data():
                    sock.sendto(data, 0, (str(self.IP4_MULTICAST), self.MULTICAST_PORT))
                nxt += delay
                delay *= 2

            r, w, e = select.select([sock], [], [sock], 0.5)
            if not r:
                continue

            data, addr = sock.recvfrom(16384)
            self.logger.debug("Received %d bytes from %s:%d", len(data), *addr)

            if data:
                resp = MDNSResponse(data)
                # We only want responses to our queries, not everyone elses!
                if not resp.is_valid:
                    continue

                if not resp.is_applicable(self.query):
                    continue

                for a in resp.answers:
                    if 'PTR' not in a:
                        continue

                    self.logger.debug("Answer: %s", a)

                    self.query.add_answer(a['qname'], a['PTR'], a['qtype'], a['qclass'], a['ttl'])
                    if a['PTR'] not in self.devices:
                        dev = {'PTR': a['PTR']}
                        for ad in resp.additional:
                            self.logger.debug("Additional Record: %s", ad)
                            for poss in ('A', 'AAAA', 'TXT', 'SRV'):
                                if poss in ad:
                                    dev[poss] = ad[poss]
                        if 'A' not in dev:
                            self.logger.info("No additional records found, so using origin IP for %s", a['PTR'])
                            dev['A'] = addr[0]
                        self.devices[a['PTR']] = dev

        sock.close()

        return len(self.devices) > 0

    def make_socket(self):
        """ Open a socket that can be used for sending and receiving multicast packets.
            If a socket cannot be created or setup then an MDNSServiceDiscoveryError is raised.
        :return: The created socket.
        """
        host = socket.inet_aton(self.interface)

        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)

            sock.setblocking(0)
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)

            sock.bind((str(self.IP4_MULTICAST), self.MULTICAST_PORT))

            sock.setsockopt(socket.SOL_IP, socket.IP_MULTICAST_TTL, struct.pack('B', self.ttl))
            sock.setsockopt(socket.SOL_IP, socket.IP_MULTICAST_IF, host)
            sock.setsockopt(socket.SOL_IP, socket.IP_MULTICAST_LOOP, 0)

        except:
            print(sys.exc_info())
            raise MDNSServiceDiscoveryError("Unable to create a suitable socket for MDNS")

        return sock



from datetime import datetime


class Headers(object):
    CRLF = b"\r\n"
    EOH = b"\r\n\r\n"

    def __init__(self, data=None, status_line=None):
        self.buffer = data or b''
        self.finished = False
        self.status_line = status_line
        self.headers = {}

    def __len__(self):
        return len(self.buffer)

    @property
    def needs_content(self):
        if self.get('content-length') is None and self.get('content-type') is None:
            return False
        return True

    def read_content(self, cntnt):
        """ Read data from a stream until we have a complete set of headers.
        :param cntnt: Stream data to read.
        :return: Number of bytes of stream that have been used.
        """
        if self.finished:
            return 0
        consumed = 0
        idx = cntnt.find(self.EOH)
        if idx >= 0:
            consumed = idx + 4
            self.buffer += cntnt[:idx]
            self.finished = True
        else:
            consumed = len(cntnt)
            olen = len(self.buffer)
            self.buffer += cntnt
            idx = self.buffer.find(self.EOH)
            if idx >= 0:
                consumed = idx - olen + 4
                self.buffer = self.buffer[:idx]
                self.finished = True
        if self.finished:
            self.parse_headers()
        return consumed

    def add_header(self, key, value):
        self.headers[key] = value

    def add_headers(self, hdr_dict):
        if hdr_dict is not None and len(hdr_dict) > 0:
            self.headers.update(hdr_dict)

    def parse_headers(self):
        """ Parse headers from a request.
        :param hdr_data: The header data, excluding the final \r\n\r\n
        :return: No
        """
        self.headers = {}
        lines = self.buffer.split(self.CRLF)
        self.status_line = lines[0].decode()
        for line in lines[1:]:
#            print(":: {}".format(line))
            if b':' in line:
                key, value = line.split(b':', 1)
                if key.decode().lower() == 'set-cookie':
                    self.headers.setdefault(key.decode(), []).append(value.strip().decode())
                else:
                    self.headers[key.decode()] = value.strip().decode()
            else:
                print("Malformed header line: {}".format(line))

    def __str__(self):
        lines = [self.status_line] if self.status_line is not None else []
        self.headers['Date'] = datetime.utcnow().strftime("%a, %d %b %Y %H:%M:%S GMT")
        for k in sorted(self.headers):
            if not isinstance(self.headers[k], list):
                lines.append("{}: {}".format(k, self.headers[k]))
            else:
                for v in self.headers[k]:
                    lines.append("{}: {}".format(k, v))

        return "\r\n".join(lines) + "\r\n\r\n"

    def get(self, key, default=None):
        """ Get a header value or default, regardless of case.
        :param key: The header key to get. Case insensitive.
        :param default: The value to return if the header s not present. Defaults to None.
        :return: The value of the header or the default value (if no default is given, None).
        """
        key = key.decode() if isinstance(key, bytes) else key
        for k in self.headers.keys():
            kk = k if not isinstance(k, bytes) else k.decode()
            if kk.lower() == key.lower():
                v = self.headers[k]
                if not isinstance(v, list):
                    return int(v) if v.isdigit() else v
                return v
        return default
